unit gamesa;

interface

type dvojice = array[0..1] of longint;
procedure herni_plan (_n : longint; uhlopricky : array of dvojice);

procedure tah_honzika ( var hrana : dvojice);

procedure tah_marenky ( hrana : dvojice);

implementation

const MAXN = 111111;

var
  N : longint;
  degree : array[0..MAXN] of longint;
  line : array[0..2*MAXN] of dvojice;
  color : array[0..2*MAXN] of boolean;
  last_marenka : longint;
  sorted_line : array[0..2*MAXN] of longint;

  _uhlopricky : array[0..MAXN] of dvojice;
  sorted_uhlopricky : array[0..MAXN] of longint;

function compare( a, b : dvojice ) : longint; begin
  compare := 0;
  if a[0] > b[0] then compare := 1
  else if a[0] < b[0] then compare := -1
  else if a[1] > b[1] then compare := 1
  else if a[1] < b[1] then compare := -1;
end;

procedure sort(var a : array of dvojice; l, r : longint; var sorted : array of longint);
var
  i, j, x, tmp: longint;
begin
  i := l; j := r;
  x := sorted[l + random(r-l+1)];
  repeat
    while compare(a[x], a[sorted[i]]) = 1 do inc(i);
    while compare(a[sorted[j]], a[x]) = 1 do dec(j);
    if i <= j then
    begin
      tmp := sorted[i]; sorted[i] := sorted[j]; sorted[j] := tmp;
      inc(i); dec(j);
    end;
  until i > j;
  if l < j then Sort(a, l, j, sorted);
  if i < r then Sort(a, i, r, sorted);
end;

function binary_search(var a : array of dvojice; var sorted : array of longint; x : dvojice; n : longint) : longint;
var
  left, right, middle : longint;
begin
  left := 0;
  right := n-1;
  while left < right do begin
    middle := (left + right) div 2;
    if compare(a[sorted[middle]], x) < 0 then left := middle + 1
    else right := middle;
  end;
  binary_search := sorted[left];
end;

function get_hrana( hrana : dvojice) : longint;
var
  tmp : longint;
  x : dvojice;
begin
  x[0] := hrana[0]; x[1] := hrana[1];
  tmp := binary_search(line, sorted_line, x, N);
  if (line[tmp, 0] = hrana[0]) and (line[tmp, 1] = hrana[1]) then get_hrana := tmp;

  x[0] := hrana[1]; x[1] := hrana[0];
  tmp := binary_search(line, sorted_line, x, N);
  if (line[tmp, 0] = hrana[1]) and (line[tmp, 1] = hrana[0]) then get_hrana := tmp;
end;

function is_uhlopricka( a, b : longint) : boolean;
var
  tmp : longint;
  x : dvojice;
begin
  is_uhlopricka := false;

  x[0] := a; x[1] := b;
  tmp := binary_search(_uhlopricky, sorted_uhlopricky, x, N - 3);
  if (_uhlopricky[tmp, 0] = a) and (_uhlopricky[tmp, 1] = b) then is_uhlopricka := true;

  x[0] := b; x[1] := a;
  tmp := binary_search(_uhlopricky, sorted_uhlopricky, x, N - 3);
  if (_uhlopricky[tmp, 0] = b) and (_uhlopricky[tmp, 1] = a) then is_uhlopricka := true;
end;

procedure error;
begin
  writeln('Chyba vstupu!');
  halt;
end;

procedure herni_plan (_n : longint; uhlopricky : array of dvojice);
var
  i, left, right, next_left, next_right, count: longint;
begin
  N := _n;

  if (N < 3) or (N > 100000) then error;

  randomize;

  for i := 0 to N-1 do degree[i] := 0;

  for i := 0 to N-4 do begin

    if (uhlopricky[i, 0] < 1) or (uhlopricky[i, 0] > N) or (uhlopricky[i, 1] < 1) or (uhlopricky[i, 1] > N) then error;

    inc(degree[uhlopricky[i,0]-1]);
    inc(degree[uhlopricky[i,1]-1]);
    _uhlopricky[i, 0] := uhlopricky[i, 0]-1;
    _uhlopricky[i, 1] := uhlopricky[i, 1]-1;
    
    if (uhlopricky[i, 0] = uhlopricky[i, 1]) or ((_uhlopricky[i, 0] + 1) mod N = _uhlopricky[i, 1]) or ((_uhlopricky[i, 1] + 1) mod N = _uhlopricky[i, 0]) then error;

    sorted_uhlopricky[i] := i;
  end;

  sort(_uhlopricky, 0, N-4, sorted_uhlopricky);

  count := 0;
  for i := 0 to N-1 do if degree[i] = 0 then begin
    left := i;
    inc(count);
  end;
  right := left;

  if (N > 3) and (count <> 2) then error;

  left := (left + 1) mod N;
  line[0, 0] := left; line[0, 1] := right;
  for i := 0 to N-4 do begin
    next_left := (left + 1) mod N;
    next_right := (right + N - 1) mod N;
    if(is_uhlopricka(left, next_right)) then begin
      line[2*i+1, 0] := right; line[2*i+1, 1] := next_right;
      right := next_right;
      line[2*i+2, 0] := right; line[2*i+2, 1] := left;
    end
    else if(is_uhlopricka(right, next_left)) then begin
      line[2*i+1, 0] := left; line[2*i+1, 1] := next_left;
      left := next_left;
      line[2*i+2, 0] := left; line[2*i+2, 1] := right;
    end
    else error;
  end;
  next_left := (left + 1) mod N;
  line[2*N-5, 0] := left; line[2*N-5, 1] := next_left;
  line[2*N-4, 0] := right; line[2*N-4, 1] := next_left;

  N := 2*N-3;
  last_marenka := -1;
  for i := 0 to N-1 do begin
    color[i] := false;
    sorted_line[i] := i;
  end;

  sort(line, 0, N-1, sorted_line);
end;

procedure tah_honzika ( var hrana : dvojice);
var
  answer : longint;
begin
  if last_marenka = -1 then answer := N div 2
  else begin
    if last_marenka mod 2 = 1 then begin
      if color[last_marenka-1] then answer := last_marenka+1
      else if color[last_marenka+1] then answer := last_marenka-1
      else answer := N - 1 - last_marenka;
    end
    else begin
      if (last_marenka > 0) and (color[last_marenka-1]) then answer := last_marenka - 2
      else if (last_marenka > 0) and (color[last_marenka-2]) then answer := last_marenka - 1
      else if (last_marenka < N-1) and (color[last_marenka+1]) then answer := last_marenka + 2
      else if (last_marenka < N-1) and (color[last_marenka+2]) then answer := last_marenka + 1
      else answer := N - 1 - last_marenka;
    end;
  end;
  
  color[answer] := true;
  hrana[0] := line[answer, 0] + 1;
  hrana[1] := line[answer, 1] + 1;
end;

procedure tah_marenky ( hrana : dvojice);
var
  id : longint;
  _hrana : dvojice;
begin
  _hrana[0] := hrana[0] - 1;
  _hrana[1] := hrana[1] - 1;
  id := get_hrana(_hrana);
  
  if ((line[id, 0] <> _hrana[0]) or (line[id, 1] <> _hrana[1])) and ((line[id, 0] <> _hrana[1]) or (line[id, 1] <> _hrana[0])) then error;

  color[id] := true;
  last_marenka := id;
end;

end.
