#include <stdio.h>
#include <stdlib.h>
#include <algorithm>
#include "gamesa.h"

#define MAXN 1000000

using namespace std;

typedef pair<int, int> EDGE;

int n;
int nEdges;
int nDiags;

EDGE Edge[MAXN]; //sorted lexicographically
int EdgeId[MAXN];

EDGE EdgeById[MAXN]; //sorted according to the polygon
int Red[MAXN];

int Degree[MAXN];
/*int CurIndex[MAXN];
int Fol[MAXN];
int iFol[MAXN];*/

int NextMove;

int Swap(int a, int swap)
{
	return (swap) ? ((nEdges - 1) - a) : a;
}

int Norm(int a)
{
	return (a + n) % n;
}

pair<int, int> MakePair(int a, int b)
{
	return (a < b) ? make_pair(a, b) : make_pair(b, a);
}

int EdgeIndex(int a, int b)
{
	return lower_bound(Edge, Edge + nEdges, MakePair(a, b)) - Edge;
}

void herni_plan(int nLocal, int uhlopricky[][2])
{
	n = nLocal;
	nDiags = n - 3;
	
	for(int i = 0; i < nDiags; i++)
	{
		Edge[nEdges++] = MakePair(uhlopricky[i][0] - 1, uhlopricky[i][1] - 1);
		Degree[uhlopricky[i][0] - 1]++;
		Degree[uhlopricky[i][1] - 1]++;
	}
	
	for(int i = 0; i < n; i++)
	{
		Edge[nEdges++] = MakePair(i, Norm(i + 1));
	}
	
	sort(Edge, Edge + nEdges);
	
	/*for(int i = 0; i < n; i++)
	{
		iFol[i + 1] = iFol[i] + Degree[i];
	}
	
	int from, to;
	for(int i = 0; i < n - 3; i++)
	{
		from = Edge[i].From;
		to = Edge[i].To;
		Fol[iFol[from] + CurIndex[from]++] = to;
	}*/
	
	int v;
	for(v = 0; Degree[v]; v++);
	
	EdgeId[EdgeIndex(v, Norm(v + 1))] = 0;
	EdgeId[EdgeIndex(v, Norm(v - 1))] = 1;
	
	int down = Norm(v - 1);
	int up = Norm(v + 1);
	int next, q;
	
	for(int i = 0; 1; i++)
	{
		EdgeId[EdgeIndex(down, up)] = (i * 2) + 2;
		
		if(i == nDiags - 1)
		{
			break;
		}
		
		if(q = EdgeIndex(Norm(down - 1), up), (q == nEdges) || (Edge[q] != MakePair(Norm(down - 1), up)))
		{
			next = Norm(up + 1);
			EdgeId[lower_bound(Edge, Edge + nEdges, MakePair(up, next)) - Edge] = (i * 2) + 3;
			up = next;
		}
		else
		{
			next = Norm(down - 1);
			EdgeId[lower_bound(Edge, Edge + nEdges, MakePair(down, next)) - Edge] = (i * 2) + 3;
			down = next;
		}
	}
	
	EdgeId[EdgeIndex(up, Norm(up + 1))] = nEdges - 2;
	EdgeId[EdgeIndex(down, Norm(down - 1))] = nEdges - 1;
	
	for(int i = 0; i < nEdges; i++)
	{
		EdgeById[EdgeId[i]] = Edge[i];
	}
	
	/*for(int i = 0; i < nEdges; i++)
	{
		printf("id %d: Edge %d -> %d\n", i, EdgeById[i].first + 1, EdgeById[i].second + 1);
	}*/
	
	NextMove = nEdges / 2;
}

void tah_honzika(int hrana[2])
{
	hrana[0] = EdgeById[NextMove].first + 1;
	hrana[1] = EdgeById[NextMove].second + 1;
	Red[NextMove] = 1;
}

void tah_marenky(int hrana[2])
{
	int id = EdgeId[EdgeIndex(Norm(hrana[0] - 1), Norm(hrana[1] - 1))];
	Red[id] = 1;
	
	int swap = 0;
	if(id > nEdges / 2)
	{
		id = Swap(id, swap = 1);
	}
	
	if(id == 0)
	{
		if(Red[Swap(1, swap)])
		{
			NextMove = Swap(2, swap);
			return;
		}
		
		if(Red[Swap(2, swap)])
		{
			NextMove = Swap(1, swap);
			return;
		}
	}
	
	else if(id == 1)
	{
		if(Red[Swap(0, swap)])
		{
			NextMove = Swap(2, swap);
			return;
		}
		
		if(Red[Swap(2, swap)])
		{
			NextMove = Swap(0, swap);
			return;
		}
	}
	
	else if(id % 2)
	{
		if(Red[Swap(id - 1, swap)])
		{
			NextMove = Swap(id + 1, swap);
			return;
		}
		
		if(Red[Swap(id + 1, swap)])
		{
			NextMove = Swap(id - 1, swap);
			return;
		}
	}
	
	else
	{
		if(Red[Swap(id - 1, swap)])
		{
			NextMove = Swap(id - 2, swap);
			return;
		}
		
		if(Red[Swap(id - 2, swap)])
		{
			NextMove = Swap(id - 1, swap);
			return;
		}
		
		if(Red[Swap(id + 1, swap)])
		{
			NextMove = Swap(id + 2, swap);
			return;
		}
		
		if(Red[Swap(id + 2, swap)])
		{
			NextMove = Swap(id + 1, swap);
			return;
		}
	}
	
	NextMove = Swap(id, !(swap));
}

/*
int main()
{
	int h[2];
	int u[3][2] =
	{
		{1, 3},
		{3, 6},
		{6, 4}
	};
	
	herni_plan(6, u);
	tah_honzika(h);
	printf("Honzik tahne %d -> %d\n", h[0], h[1]);
	tah_marenky((const int[2]) {2, 3});
	tah_honzika(h);
	printf("Honzik tahne %d -> %d\n", h[0], h[1]);
	tah_marenky((const int[2]) {1, 2});
	tah_honzika(h);
	printf("Honzik tahne %d -> %d\n", h[0], h[1]);
	
	return 0;
}
*/
