#include "gamesa.h"
#include <cstdio>
#include <cstdlib>
#include <unordered_map>
#include <unordered_set>
#define MAXNOPT 15
#define MAXN 100000
using namespace std;

struct stt
{
  unsigned long long val;

  stt(unsigned long long x)
    {
      val = x;
    }

  stt(void)
    {
      val = 0;
    }

  bool operator==(const stt &s) const
    {
      return val == s.val;
    }
};

namespace std {
template <> class hash< stt > : public unary_function< stt, size_t >
{
  public:
  size_t operator()(const stt &h) const
    {
      return h.val % 4294967291;
    }
};}


struct hrana
{
  unsigned vs[2];

  hrana(void)
    {
      vs[0] = vs[1] = 0;
    }

  hrana(int v1, int v2)
    {
      if (v1 < v2)
	{
	  vs[0] = v1;
	  vs[1] = v2;
	}
      else
	{
	  vs[0] = v2;
	  vs[1] = v1;
	}
    }

  bool operator==(const hrana &s) const
    {
      return vs[0] == s.vs[0] && vs[1] == s.vs[1];
    }
};

static unsigned n;
namespace std {
template <> class hash< hrana > : public unary_function< hrana, size_t >
{
  public:
  size_t operator()(const hrana &h) const
    {
      return h.vs[0] * n + h.vs[1];
    }
};}

static stt state;
static int troj[2 * MAXN][2][2];
static int idtohr[2 * MAXN][2];
static unordered_map<hrana,int> hrany;
static int maxobt_val[2 * MAXN - 3];
static unordered_set<int> maxobt[3];

static void
pridej_hranu (int f, int t)
{
  static int aid = 0;
  hrana h(f, t);

  hrany[h] = ++aid;
  idtohr[aid][0] = f;
  idtohr[aid][1] = t;
}

static int
najdi_hranu (int f, int t)
{
  hrana h(f,t);

  if (hrany.count (h) == 0)
    return 0;
  return hrany[h];
}

static void
vloz_trojuhelnik (int h1, int h2, int h3)
{
  int tmp, i;

  if (h2 > h3)
    {
      tmp = h2;
      h2 = h3;
      h3 = tmp;
    }

  for (i = 0; i < 2; i++)
    {
      if (!troj[h1][i][0])
	break;
      if (troj[h1][i][0] == h2 && troj[h1][i][1] == h3)
	return;
    }
  troj[h1][i][0] = h2;
  troj[h1][i][1] = h3;
}

static void
zkus_trojuhelnik (int v1, int v2, int v3)
{
  int h1 = najdi_hranu (v1, v2);
  int h2 = najdi_hranu (v2, v3);
  int h3 = najdi_hranu (v3, v1);

  if (!h1 || !h2 || !h3)
    return;

  vloz_trojuhelnik (h1, h2, h3);
  vloz_trojuhelnik (h2, h3, h1);
  vloz_trojuhelnik (h3, h1, h2);
}

static unsigned
pred (unsigned k)
{
  return k == 1 ? n : k - 1;
}

static unsigned
nasl (unsigned k)
{
  return k == n ? 1 : k + 1;
}

static void
najdi_trojuhelniky (int f, int t)
{
  zkus_trojuhelnik (f, t, pred (f));
  zkus_trojuhelnik (f, t, nasl (f));
  zkus_trojuhelnik (f, t, pred (t));
  zkus_trojuhelnik (f, t, nasl (t));
}

static int
obarveny (stt st, int h)
{
  return (st.val >> (h - 1)) & 1;
}

static int
v_trojuhelniku (stt st, int h)
{
  int i;

  for (i = 0; i < 2 && troj[h][i][0]; i++)
    if (obarveny (st, troj[h][i][0]) && obarveny (st, troj[h][i][1]))
      return 1;

  return 0;
}

static unordered_map<stt, int> ccon;

static int
get_cached (stt st)
{
  if (ccon.count (st) == 0)
    return -1;
  return ccon[st];
}

static void
cache (stt st, int h)
{
  ccon[st] = h;
}

static int
get_move (stt st)
{
  int nt = 2 * n - 3, h;
  stt nst;

  h = get_cached (st);
  if (h >= 0)
    return h;

  for (h = 1; h <= nt; h++)
    {
      if (obarveny (st, h))
	continue;

      nst = stt (st.val | 1ull << (h - 1));
      if (v_trojuhelniku (nst, h))
	{
	  cache (st, h);
	  return h;
	}
    }
  for (h = 1; h <= nt; h++)
    {
      if (obarveny (st, h))
	continue;

      nst = stt (st.val | 1ull << (h - 1));
      if (get_move (nst) == 0)
	{
	  cache (st, h);
	  return h;
	}
    }

  cache (st, 0);
  return 0;
}

static void
set_maxobt (int h, int val)
{
  maxobt[maxobt_val[h - 1]].erase (h);
  maxobt_val[h - 1] = val;
  if (val >= 0)
    maxobt[val].insert (h);
}

static int
det_obt (int h1, int h2)
{
  return (maxobt_val[h1 - 1] < 0) + (maxobt_val[h2 - 1] < 0);
}

static int
det_maxobt (int h)
{
  int amax = 0, i, v;

  for (i = 0; i < 2 && troj[h][i][0]; i++)
    {
      v = det_obt (troj[h][i][0], troj[h][i][1]);
      if (v > amax)
	amax = v;
    }

  return amax;
}

static void
recalc_maxobt (int h)
{
  if (maxobt_val[h - 1] < 0)
    return;
  set_maxobt (h, det_maxobt (h));
}

static void
record_move (int h)
{
  int i;

  set_maxobt (h, -1);

  for (i = 0; i < 2 && troj[h][i][0]; i++)
    {
      recalc_maxobt (troj[h][i][0]);
      recalc_maxobt (troj[h][i][1]);
    }

  if (n <= MAXNOPT)
    state.val |= 1ull << (h - 1);
}

static int
safe_move (void)
{
  if (!maxobt[2].empty ())
    return *maxobt[2].begin ();
  if (!maxobt[0].empty ())
    return *maxobt[0].begin ();
  return *maxobt[1].begin ();
}

void herni_plan (int an, int uhl[][2])
{
  unsigned i;

  n = an;

  for (i = 0; i < n - 1; i++)
    pridej_hranu (i + 1, i + 2);
  pridej_hranu (1, n);
  for (i = 0; i < n - 3; i++)
    pridej_hranu (uhl[i][0], uhl[i][1]);

  for (i = 0; i < n - 3; i++)
    najdi_trojuhelniky (uhl[i][0], uhl[i][1]);

  for (i = 0; i < 2 * n - 3; i++)
    {
      maxobt_val[i] = 0;
      maxobt[0].insert (i + 1);
    }
}

void tah_honzika (int tah[2])
{
  int h;
  
  if (n <= MAXNOPT)
    h = get_move (state);
  else
    h = safe_move ();

  record_move (h);
  tah[0] = idtohr[h][0];
  tah[1] = idtohr[h][1];

}

void
tah_marenky (int tah[2])
{
  int h = najdi_hranu (tah[0], tah[1]);
  record_move (h);
}
