#include <stdio.h>
#include <stdlib.h>

static int n;
static int deg[100000];

static void
edge (FILE *f, int a, int b)
{
  fprintf (f, " %d %d", a, b);
  deg[a]++;
  deg[b]++;
}

static void
genbip (FILE *f, int lft, int rt)
{
  int i, j;

  n = lft + rt;
  fprintf (f, "%d %d\n", n, lft * rt);
  for (i = 0; i < n; i++)
    deg[i] = 0;
  for (i = 0; i < lft; i++)
    for (j = lft; j < n; j++)
      edge (f, i, j);
  fprintf (f, "\n");
}

static void
genbippath (FILE *f, int lft, int rt)
{
  int i, j, m;

  n = lft + rt;
  m = lft * rt + rt - 1;
  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;
  for (i = 0; i < lft; i++)
    for (j = lft; j < n; j++)
      edge (f, i, j);
  for (j = lft; j < n - 1; j++)
    edge (f, j, j + 1);
  fprintf (f, "\n");
}

static int choice[600000];

static void
genqueries (FILE *f, int nq, const char *p)
{
  int i, j, m;

  for (i = 0; i < n; i++)
    fprintf (f, "%d ", rand () % 5);
  fprintf (f, "\n");

  m = 0;
  for (i = 0; i < n; i++)
    {
      j = m;
      m += deg[i];
      for (; j < m; j++)
	choice[j] = i;
    }

  for (i = 0; i < nq / 2; i++)
    fprintf (f, "N %d %d\n%s", choice[rand () % m], rand () % 5, p);
}

static void
gen3 (int vs[3], int rng)
{
  int i, j;

  for (i = 0; i < 3; i++)
    {
again:
      vs[i] = rand () % rng;
      for (j = 0; j < i; j++)
	if (vs[i] == vs[j])
	  goto again;
    }
}

static int
sel (int rng)
{
  int v;

  for (v = 0; v < rng - 1; v++)
    if (rand () % 3 == 0)
      return v;
  return rng - 1;
}

static void
gen3m (int vs[3], int rng)
{
  int i, j;

  for (i = 0; i < 3; i++)
    {
again:

      vs[i] = sel (rng);
      for (j = 0; j < i; j++)
	if (vs[i] == vs[j])
	  goto again;
    }
}


static void
gendlayer (FILE *f, int flay, int slay)
{
  int m = 3 * flay + 3 * (slay - 3);
  int i, j;

  n = flay + slay;

  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;

  for (i = 0; i < 3; i++)
    for (j = 3; j < slay; j++)
      edge (f, i, j);
  fprintf (f, "\n");
  for (i = slay; i < n; i++)
    {
      int vs[3];
      gen3 (vs, slay);
      for (j = 0; j < 3; j++)
	edge (f, i, vs[j]);
    }
  fprintf (f, "\n");
}

static void
gendlayerm (FILE *f, int flay, int slay)
{
  int m = 3 * flay + 3 * (slay - 3);
  int i, j;

  n = flay + slay;

  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;

  for (i = 0; i < 3; i++)
    for (j = 3; j < slay; j++)
      edge (f, i, j);
  fprintf (f, "\n");
  for (i = slay; i < n; i++)
    {
      int vs[3];
      gen3m (vs, slay);
      for (j = 0; j < 3; j++)
	edge (f, i, vs[j]);
    }
  fprintf (f, "\n");
}

static void
genslayer (FILE *f, int lft, int rt)
{
  int m = 3 * rt;
  int i, j;

  n = lft + rt;

  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;

  for (i = lft; i < n; i++)
    {
      int vs[3];
      gen3 (vs, lft);
      for (j = 0; j < 3; j++)
	edge (f, i, vs[j]);
    }
  fprintf (f, "\n");
}

static void
genslayerm (FILE *f, int lft, int rt)
{
  int m = 3 * rt;
  int i, j;

  n = lft + rt;

  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;

  for (i = lft; i < n; i++)
    {
      int vs[3];
      gen3m (vs, lft);
      for (j = 0; j < 3; j++)
	edge (f, i, vs[j]);
    }
  fprintf (f, "\n");
}

static int perm[100000];

static void
swap (int i, int j)
{
  int t = perm[i];
  perm[i] = perm[j];
  perm[j] = t;
}

static void
genranddeg (FILE *f, int nv)
{
  int i, j;
  int m;

  n = nv;
  m = 3 * (n-3);
  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    {
      deg[i] = 0;
      perm[i] = i;
    }
  for (i = 0; i < n; i++)
    swap (i, i + rand () % (n - i));
  for (i = 3; i < n; i++)
    {
      int vs[3];
      gen3m (vs, i);
      for (j = 0; j < 3; j++)
	edge (f, perm[i], perm[vs[j]]);
    }
  fprintf (f, "\n");
}

static int
pnt (int a, int b)
{
  return 100 * a + b;
}

static void
gengrid (FILE *f)
{
  int m;
  int i, j;

  n = 10000;
  m = 2 * 100 * 99 + 99 * 99;
  fprintf (f, "%d %d\n", n, m);
  for (i = 0; i < n; i++)
    deg[i] = 0;
  for (i = 0; i < 100; i++)
    for (j = 0; j < 99; j++)
      {
	edge (f, pnt(i,j), pnt(i, j+1));
	edge (f, pnt(j,i), pnt(j+1, i));
      }
  for (i = 0; i < 99; i++)
    for (j = 0; j < 99; j++)
      edge (f, pnt(i,j), pnt(i+1, j+1));
  fprintf(f, "\n");
}

int main(void)
{
  FILE *f;
  int i;

  srand (0);

  /* 2 body za n <= 100.  */
  f = fopen ("01.in", "w");
  fprintf (f, "4 4\n");
  fprintf (f, "0 1 1 2 2 0 2 3\n");
  fprintf (f, "0 0 1 0\n");
  fprintf (f, "P\nN 2 0\nP\nN 3 1\nP\nN 2 1\nP\nN 0 1\nN 1 1\nP\n");
  fclose (f);

  f = fopen ("02.in", "w");
  genbip (f, 3, 97);
  genqueries (f, 1000, "P\n");
  fclose (f);

  /* 2 body za jednou volane P.  */
  f = fopen ("03.in", "w");
  gendlayer (f, 20000, 10);
  for (i = 0; i < n; i++)
    fprintf (f, "%d ", rand () % 5);
  fprintf (f, "\n");
  fprintf (f, "P\n");
  fclose (f);

  f = fopen ("04.in", "w");
  genranddeg (f, 100000);
  genqueries (f, 100000, "");
  fprintf (f, "P\n");
  fclose (f);

  /* 5 bodu za zadne hrany mezi velkymi.  */
  f = fopen ("05.in", "w");
  gengrid (f);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("06.in", "w");
  genbip (f, 3, 99997);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("07.in", "w");
  genslayer (f, 10, 10000);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("08.in", "w");
  genslayer (f, 100, 99900);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("09.in", "w");
  genslayerm (f, 100, 99900);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  /* 6 bodu za zbytek.  */
  f = fopen ("10.in", "w");
  genbippath (f, 2, 99998);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("11.in", "w");
  gendlayer (f, 50000, 1000);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("12.in", "w");
  gendlayer (f, 99900, 100);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("13.in", "w");
  genranddeg (f, 100000);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("14.in", "w");
  gendlayerm (f, 99900, 100);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  f = fopen ("15.in", "w");
  gendlayerm (f, 50000, 1000);
  genqueries (f, 1000000, "P\n");
  fclose (f);

  return 0;
}
